--- Main implementation file
--
--  BigRed maintains an array of instances of the Message class, which is a
--  wrapper around a FontString with some other goodies added in. Whenever
--  the maximum number of messages is increased, new instances of Message
--  are created to accommodate the new slots. They are not freed, even if the
--  max number of messages is later decreased.

local _, private = ...;

local OnEnable_RunOnce, EvictLRU, DrawMessage, RemoveMessage, SlashCmd, OnUpdate, CheckUpdating, OnEvent,
UpdateMessageFont;
local Message;

local frame, anchor;
local db;

BigRed = LibStub("AceAddon-3.0"):NewAddon("BigRed");

-- number of msgs currently visible
local size = 0;

-- list of messages
local messages = {};

-- map from text displayed to message instance
-- note that when a message is completely faded, it is no longer represented in
-- this map
local textToMessage = {};

-- frame anchors wrt text justification
private.anchors = {
  LEFT = "TOPLEFT",
  CENTER = "TOP",
  RIGHT = "TOPRIGHT",
};

function BigRed:OnEnable ()
  OnEnable_RunOnce(self);
  
  frame:Show();
  if not db.profile.isLocked then
    anchor:Show();
  end
  
  frame:RegisterEvent("UI_ERROR_MESSAGE");
  UIErrorsFrame:UnregisterEvent("UI_ERROR_MESSAGE");

  SLASH_BIGRED1 = "/bigred";
  SlashCmdList["BIGRED"] = SlashCmd;

  frame:SetScript("OnEvent", OnEvent);
  private.ApplySettings();
end

local HasRun = false;

function OnEnable_RunOnce ()
  if HasRun then
    return;
  else
    HasRun = true;
  end
  
  private.db = LibStub("AceDB-3.0"):New("BigRedDB", private.GetDefaults());
  db = private.db;

  db.RegisterCallback(BigRed, "OnProfileChanged", private.ApplySettings);
  db.RegisterCallback(BigRed, "OnProfileCopied", private.ApplySettings);
  db.RegisterCallback(BigRed, "OnProfileReset", private.ApplySettings);

  private.frame = LibStub("LuaToGui"):Create(private.frameTable, UIParent);
  private.anchor = LibStub("LuaToGui"):Create(private.anchorTable);

  frame, anchor = private.frame, private.anchor;
  

  -- background texture
  do
    local t = anchor:CreateTexture();
    t:SetTexture(0,0,0,0.4);
    t:SetAllPoints(anchor);
  end
  
  private.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(db);
  LibStub("AceConfig-3.0"):RegisterOptionsTable("BigRed", private.options);
  
end

function BigRed:OnDisable ()
  frame:Hide();
  anchor:Hide();
  frame:UnregisterEvent("UI_ERROR_MESSAGE");
  UIErrorsFrame:RegisterEvent("UI_ERROR_MESSAGE");

  _G.SLASH_BIGRED1 = nil;
  _G.SlachCmdList["BIGRED"] = nil;

  frame:SetScript("OnEvent", nil);
end

-- adds the message 'text' to BigRed's display, as if it were an error message
function BigRed:AddMessage (text)
  private.AssertType("message", text, "string");
  local message = textToMessage[text];
  
  -- add a new message
  if (nil == message) or (not message:IsActive()) then
    if size == db.profile.MaxMessages then
      message = EvictLRU();
    else
      local cap = db.profile.MaxMessages;
      -- find an empty spot
      for i = 1, cap do
        if not messages[i]:IsActive() then
          message = messages[i];
          break;
        end
      end
    end

    textToMessage[text] = message;
    DrawMessage(text, message);
  end
  
  message:SetTime(GetTime());
  CheckUpdating();
end

local numFontStrings = 0;

function private.UpdateMaxMessages (newMax)
  if newMax > numFontStrings then
    for i = (numFontStrings + 1), newMax do
      messages[i] = private.Message:New(i);
      UpdateMessageFont(messages[i]);
    end
    numFontStrings = newMax;
  elseif newMax < numFontStrings then
    for i = (newMax + 1), numFontStrings do
      if messages[i]:IsActive() then
        RemoveMessage(messages[i]);
      end
    end
  end
end

-- evicts the least recently used message, returning the index freed
function EvictLRU ()
  local _, evictee = next(textToMessage);
  
  for _, message in ipairs(textToMessage) do
    if evictee.time > message.time then
      evictee = message;
    end 
  end

  RemoveMessage(evictee);
  return evictee;
end

function DrawMessage (text, message)
  message:MakeActive();
  message:SetText(text);
  size = size + 1;
end

function RemoveMessage (message)
  size = size - 1;
  message:MakeInactive();
  textToMessage[message.text] = nil;
  message.text = nil;
end

function private.OnDragStart ()
  frame:StartMoving();
end

function private.OnDragStop ()
  frame:StopMovingOrSizing();
  db.profile.x, db.profile.y = BigRed:GetXY();
  BigRed:SetXY(db.profile.x, db.profile.y)
end

function SlashCmd ()
  LibStub("AceConfigDialog-3.0"):Open("BigRed");
end

function private.AssertType (...)
  -- LibStub("-- Assert-1.0"):Type(...);
end

function  private.AssertBounds (...)
  -- LibStub("-- Assert-1.0"):Bounds(...);
end

-- Applies all of BigRed's settings anew.
function private.ApplySettings ()
  BigRed:SetXY(db.profile.x, db.profile.y);
  BigRed:SetMaxMessages(db.profile.MaxMessages);
  BigRed:SetWidth(db.profile.Width);
  BigRed:SetScale(db.profile.Scale);
  BigRed:SetFadePercent(db.profile.FadePercent);
  BigRed:SetFadeTime(db.profile.FadeTime);
  BigRed:SetFont(db.profile.Font);
  BigRed:SetFontOutline(db.profile.FontOutline);
  BigRed:SetFontJustify(db.profile.FontJustify);
  BigRed:SetLocked(db.profile.isLocked);
end

do
  local upt = 0
  
  function OnUpdate (_, t)
    upt = t + upt;
    if upt > private.UPDATE_RATE then
      upt = 0;
      local curr_time = GetTime();

      -- ping all live messages
      for i=1, db.profile.MaxMessages do
        if messages[i]:IsActive() then
          -- if it's dead after this update, remove it
          if messages[i]:Update(curr_time) then
            RemoveMessage(messages[i]);
          end
        end
      end

      CheckUpdating();
    end
  end

end

-- Checks whether we need calls to OnUpdate
function CheckUpdating ()
  for _, message in ipairs(messages) do
    if message:IsActive() then
      frame:SetScript("OnUpdate", OnUpdate);
      return;
    end
  end

  frame:SetScript("OnUpdate", nil);
end

function OnEvent (_, event, msg)
  BigRed:AddMessage(msg);
end

function private.UpdateAllFonts ()
  for _, message in ipairs(messages) do
    UpdateMessageFont(message);
  end
end

function UpdateMessageFont (message)
  local font = LibStub("LibSharedMedia-3.0"):Fetch("font", db.profile.Font);
  message:SetFont(font, db.profile.FontOutline, db.profile.FontJustify);
end

-- does NOT read db.profile
function BigRed:GetXY ()
  local x, y;
  
  if db.profile.FontJustify == "LEFT" then
    x = frame:GetLeft();
  elseif db.profile.FontJustify == "RIGHT" then
    x = frame:GetRight();
  else
    x = frame:GetLeft() + (db.profile.Width / 2);
  end

  y = frame:GetTop();
  return x, y;
end

-- does NOT update db
function BigRed:SetXY (x, y)
  frame:ClearAllPoints();
  frame:SetPoint(private.anchors[db.profile.FontJustify],
                 UIParent,
                 "BOTTOMLEFT",
                 x,
                 y);
end

private.Message = {};

private.Message.metatable = { __index = private.Message };

function private.Message:New (position)
  local m = setmetatable({}, private.Message.metatable);
  m.position = position;

  local offset = -((m.position - 1) * private.FONT_HEIGHT);
  m.fontstring = frame:CreateFontString();
  m.fontstring:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, offset);
  m.fontstring:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, offset);
  m.fontstring:SetShadowColor(0, 0, 0, 1);
  m.fontstring:SetShadowOffset(1, -1);

  m:MakeInactive();
  
  return m;
end

function private.Message:IsActive ()
  return self.active;
end

function private.Message:MakeActive ()
  self.active = true;
  self.fontstring:Show();
end

function private.Message:MakeInactive ()
  self.active = false;
  self.fontstring:Hide();
end

-- returns true if the message is done fading
function private.Message:Update (curr_time)
  local p = (curr_time - self.time) / db.profile.FadeTime;
  local r, g, b = db.profile.r, db.profile.g, db.profile.b;

  if p < private.FLASH_DURATION then
    -- flash it bright
    self.fontstring:SetTextColor(1.0, 1.0, 1.0, 1);
  elseif p < (1 - db.profile.FadePercent) then
    -- let it be constant
    self.fontstring:SetTextColor(r, g, b, 1);
  else
    -- fade it out
    local percentFaded = p - (1 - db.profile.FadePercent);
    local a = 1 - (percentFaded / db.profile.FadePercent);
    self.fontstring:SetTextColor(r, g, b, a);
    if a <= 0 then
      return true;
    end
  end
end

function private.Message:SetTime (time)
  self.time = time;
end

function private.Message:SetText (text)
  self.text = text;
  self.fontstring:SetText(text);
end

function private.Message:SetFont (font, outline, justify)
  self.fontstring:SetFont(font, private.FONT_HEIGHT, outline);
  self.fontstring:SetJustifyH(justify);
end

