--- LuaToGui-0.1 library.
--
--  Provides tools for creating GUI elements from Lua tables that describe
--  them.

local MAJOR, MINOR = "LuaToGui", 3;
LuaToGui = LibStub:NewLibrary(MAJOR, MINOR);

if not LuaToGui then
  return;
end

local convertElement, getFieldWithContext, buildRegion, handleScripts;

local handlers = {};

function LuaToGui:Create (t, parentFrame)
  local context = Context:new();
  context:setParentFrame(parentFrame);
  return convertElement(t, context);
end

function LuaToGui:AddHandler (type, handler)
  if handlers[type] then
    error(("handler of type '%s' already exists"):format(type));
  else
    handlers[type] = handler;
  end
end

local function ApplyField (element, applyFunc, ...)
  if (...) then
    return applyFunc(element, ...);
  end
end

function convertElement (t, context)
  return handlers[t.type](t, context);
end

local function ApplyRegisterForDrag (element, buttons)
  for _, button in ipairs(buttons) do
    element:RegisterForDrag(button);
  end
end

local function ApplyChildren (element, children)
  -- create context in which this is the parent
  local newContext = Context:new();
  newContext:setParentFrame(element);
  
  -- must be ipairs so that children are instantiated in the right order
  for _,v in ipairs(children) do
    convertElement(v, newContext);
  end
end

function handlers.frame (t, context)
  local parent = context:getParentFrame();
  local frame = CreateFrame(t.type, t.name, parent, t.inheritsFrom);
  
  buildRegion(t, context, frame);
  handleScripts(t, frame);

  ApplyField(frame, frame.Hide, t.hidden);
  ApplyField(frame, frame.SetBackdrop, t.backdrop);
  ApplyField(frame, frame.SetClampedToScreen, t.clampedToScreen);
  ApplyField(frame, frame.EnableMouse, t.EnableMouse);
  ApplyField(frame, frame.SetFrameStrata, t.frameStrata);
  ApplyField(frame, frame.SetFrameLevel, t.frameLevel);
  ApplyField(frame, frame.SetMovable, t.movable);
  ApplyField(frame, ApplyRegisterForDrag, t.registerForDrag);
  ApplyField(frame, ApplyChildren, t.children);

  return frame;
end

function handlers.button (t, context)
  local button = handlers.frame(t, context);

  ApplyField(button, button.SetNormalFontObject, t.normalFont);
  ApplyField(button, button.SetHighlightFontObject, t.highlightFont);

  if t.textJustifyH then
    local fontObj = button:GetNormalFontObject();
    fontObj:SetJustifyH(t.textJustifyH);
  end
  
  ApplyField(button, button.SetNormalTexture, t.normalTexture);
  ApplyField(button, button.SetPushedTexture, t.pushedTexture);
  ApplyField(button, button.RegisterForClicks, unpack(t.registerForClicks or {}));
  
  if t.highlightTexture then
    button:SetHighlightTexture(t.highlightTexture, t.highlightMode);
    if t.highlightTextureColor then
      local highlightTexture = button:GetHighlightTexture();
      highlightTexture:SetVertexColor(t.highlightTextureColor.r,
                                      t.highlightTextureColor.g,
                                      t.highlightTextureColor.b,
                                      t.highlightTextureColor.a);
    end
  end

  ApplyField(button, button.SetDisabledTexture, t.disabledTexture);
  
  return button;
end

function handlers.texture (t, context)
  local parent = context:getParentFrame();
  local name = getFieldWithContext(t, "name", context);
  
  local texture = parent:CreateTexture(name, t.drawLayer, t.inheritsFrom);
  
  if t.texture then
    texture:SetTexture(t.texture);
    if t.color then
      texture:SetVertexColor(t.color.r, t.color.g, t.color.b, t.color.a);
    end
  elseif t.color then
    texture:SetTexture(t.color.r, t.color.g, t.color.b, t.color.a);
  end
  
  if t.texCoords then
    local tc = t.texCoords;
    texture:SetTexCoord(tc.left, tc.right, tc.top, tc.bottom);
  end

  ApplyField(texture, texture.SetBlendMode, t.blendMode);

  buildRegion(t, context, texture);
  
  return texture;
end

function handlers.fontstring (t, context)
  local parent = context:getParentFrame();
  local name = getFieldWithContext(t, "name", context);
  
  local fontstring = parent:CreateFontString(name, t.drawLayer, t.inheritsFrom);

  ApplyField(fontstring, fontstring.SetFont, t.font, t.fontHeight);
  ApplyField(fontstring, fontstring.SetTextHeight, t.fontHeight);
  
  ApplyField(fontstring, fontstring.SetText, t.text);

  ApplyField(fontstring, fontstring.SetJustifyV, t.justifyV);
  ApplyField(fontstring, fontstring.SetJustifyH, t.justifyH);

  buildRegion(t, context, fontstring);
  
  return fontstring;
end

function getFieldWithContext (table, key, context)
  local value = table[key];
  if "string" == type(value) then
    if value:find("$parent") then
      value = value:gsub("$parent", context:getParentFrame():GetName());
    end
    if "!" == value:sub(1,1) then
      value = _G[value:sub(2)];
    end
  end
  
  return value;
end

function buildRegion (t, context, region)
  if t.anchors then
    if t.anchors.allPoints then
      region:SetAllPoints(getFieldWithContext(t.anchors, "allPoints", context));
    else
      for point, args in pairs(t.anchors) do
        point = point or args.point;
        local relativeTo = getFieldWithContext(args, "relativeTo", context) or
                           context:getParentFrame();
        local relativePoint = args.relativePoint or point;
        region:SetPoint(point, relativeTo, relativePoint,
                        args.xOff, args.yOff);
      end
    end
  end
  
  ApplyField(region, region.SetWidth, t.width);
  ApplyField(region, region.SetHeight, t.height);
  ApplyField(region, region.SetScale, t.scale);
end

function handleScripts (t, element)
  if t.scripts then
    for handlerName, func in pairs(t.scripts) do
      element:SetScript(handlerName, func);
    end
  end
end
