local _, scope = ...;
setmetatable(scope, { __index = _G });
setfenv(1, scope);

SCALE_MIN = 0.3;
SCALE_MAX = 2.0;
WIDTH_MIN = 100;
WIDTH_MAX = 1000;
MAX_MESSAGES_MIN = 1;
MAX_MESSAGES_MAX = 10;
FADE_PERCENT_MIN = 0;
FADE_PERCENT_MAX = 1.0;
FADE_TIME_MIN = 1;
FADE_TIME_MAX = 15;
R_MIN, G_MIN, B_MIN = 0, 0, 0;
R_MAX, G_MAX, B_MAX = 0, 0, 0;

FONT_OUTLINE_VALUES = {
  NONE = "None",
  THINOUTLINE = "Thin",
  THICKOUTLINE = "Thick",
};

FONT_JUSTIFY_VALUES = {
  LEFT = "Left",
  CENTER = "Center",
  RIGHT = "Right",
};

-- defaults
function GetDefaults ()
  _G.scale = UIParent:GetScale();
  return  {
    profile = {
      isLocked = true,
      MaxMessages = 3,
      Scale = 1,
      Width = 400,
      FadePercent = 0.5,
      FadeTime = 5,
      Font = LibStub("LibSharedMedia-3.0"):GetDefault("font"),
      FontOutline = "NONE",
      FontJustify = "CENTER",
      x = (UIParent:GetRight()/2),
      y = (UIParent:GetTop() - 122),
      r = 0.75,
      g = 0,
      b = 0,
    }
  };
end

function BigRed:SetScale (scale)
  AssertBounds("scale", scale, SCALE_MIN, SCALE_MAX);
  
  local old_scale = db.profile.Scale;
  db.profile.Scale = scale;
  frame:SetScale(scale);

  local old_x, old_y = BigRed:GetXY();
  local new_x = old_x * (old_scale / scale);
  local new_y = old_y * (old_scale / scale);
  db.profile.x, db.profile.y = new_x, new_y;

  BigRed:SetXY(db.profile.x, db.profile.y);
end

function BigRed:GetScale ()
  return db.profile.Scale;
end

function BigRed:SetWidth (width)
  AssertBounds("width", width, WIDTH_MIN, WIDTH_MAX);

  db.profile.Width = width;
  frame:SetWidth(width);
end

function BigRed:GetWidth ()
  return db.profile.Width;
end

function BigRed:SetMaxMessages (max)
  AssertBounds("max", max, MAX_MESSAGES_MIN, MAX_MESSAGES_MAX);

  UpdateMaxMessages(max);
  db.profile.MaxMessages = max;
  frame:SetHeight(max * FONT_HEIGHT);
end

function BigRed:GetMaxMessages ()
  return db.profile.MaxMessages;
end

function BigRed:SetFadePercent (n)
  AssertType("n", n, "number");
  AssertBounds("n", n, FADE_PERCENT_MIN, FADE_PERCENT_MAX);
  
  db.profile.FadePercent = n;
end

function BigRed:GetFadePercent ()
  return db.profile.FadePercent;
end

function BigRed:SetFadeTime (n)
  AssertType("n", n, "number");
  AssertBounds("n", n, FADE_TIME_MIN, FADE_TIME_MAX);
  
  db.profile.FadeTime = n;
end

function BigRed:GetFadeTime ()
  return db.profile.FadeTime;
end

function BigRed:SetFont (font)
  AssertType("font", font, "string");
  
  db.profile.Font = font;
  UpdateAllFonts();
end

function BigRed:GetFont ()
  return db.profile.Font;
end
  
function BigRed:SetFontOutline (outline)
  AssertType("outline", outline, "string");
  
  db.profile.FontOutline = outline;
  UpdateAllFonts();
end

function BigRed:GetFontOutline ()
  return db.profile.FontOutline;
end

function BigRed:SetFontJustify (justify)
  AssertType("justify", justify, "string");
  db.profile.FontJustify = justify;
  local new_x, new_y = BigRed:GetXY();
  
  frame:ClearAllPoints();
  frame:SetPoint(anchors[justify], UIParent, "BOTTOMLEFT", new_x, new_y);
  
  db.profile.x = new_x;
  db.profile.y = new_y;
  UpdateAllFonts();
end

function BigRed:GetFontJustify()
  return db.profile.FontJustify;
end
  
function BigRed:SetLocked (state)
  AssertType("state", state, "boolean");
  
  db.profile.isLocked = state
  if state then
    anchor:Hide();
  else
    anchor:Show();
  end
end

function BigRed:GetLocked ()
  return db.profile.isLocked;
end    

function BigRed:SetColor (r, g, b)
  AssertBounds("r", r, R_MIN, R_MAX);
  AssertBounds("g", g, G_MAX, G_MAX);
  AssertBounds("b", b, B_MAX, B_MAX);
  
  local s = db.profile;
  s.r, s.g, s.b = r, g, b;
end

function BigRed:GetColor ()
  local s = db.profile;
  return s.r, s.g, s.b;
end

-- Base of the options table
options = {
  type = "group",
    args = {
      general = {
        name = "General",
        desc = "Configure BigRed's behavior.",
        type = "group",
        args = {
          scale = {
            type = "range",
            name = "Scale",
            desc = "Set BigRed's scaling",
            get = BigRed.GetScale,
            set = BigRed.SetScale,
            min = SCALE_MIN,
            max = SCALE_MAX,
            step = 0.1, bigStep = 0.1,
          },
          width = {
            type = "range",
            name = "Width",
            desc = "Set BigRed's width",
            get = BigRed.GetWidth,
            set = BigRed.SetWidth,
            min = WIDTH_MIN,
            max = WIDTH_MAX,
            step = 1, bigStep = 10,
          },
          max_messages = {
            type = "range",
            name = "Max Messages",
            desc = "Set the maximum number of messages to display at once",
            get = BigRed.GetMaxMessages,
            set = BigRed.SetMaxMessages,
            min = MAX_MESSAGES_MIN,
            max = MAX_MESSAGES_MAX,
            step = 1, bigStep = 1,
          },
          font = {
            name = "Font",
            type = "select",
            desc = "Select BigRed's font",
            dialogControl = 'LSM30_Font',
            set = BigRed.SetFont,
            get = BigRed.GetFont,
            style = "dropdown",
            values = AceGUIWidgetLSMlists.font,
          },
          font_outline = {
            name = "Font outline",
            type = "select",
            desc = "Select font outline thickness",
            set = BigRed.SetFontOutline,
            get = BigRed.GetFontOutline,
            style = "dropdown",
            values = FONT_OUTLINE_VALUES,
          },
          font_justify = {
            name = "Text justification",
            type = "select",
            desc = "Select text justification",
            set = BigRed.SetFontJustify,
            get = BigRed.GetFontJustify,
            style = "dropdown",
            values = FONT_JUSTIFY_VALUES,
          },
          lock = {
            name = "Lock",
            desc = "Toggle the addon's lock",
            type = "toggle",
            get = BigRed.GetLocked,
            set = BigRed.SetLocked,
          },
          color = {
            name = "Color",
            type = "color",
            desc = "Set text color",
            set = BigRed.SetColor,
            get = BigRed.GetColor,
          },
        },
      },
    },
};